"""diagnostics contains routines for creating extracting key
features from a set of cells"""

import numpy as np
import pandas as pd

from cellpy import prms
from cellpy.exceptions import NullData


def first_cycle_irreversible_capacity():
    ...
