from cellpy.utils.batch_tools.batch_core import BaseReporter


class HTMLReporter(BaseReporter):
    def __init__(self):
        super().__init__()


class PPTReporter(BaseReporter):
    def __init__(self):
        super().__init__()
